@@ -30,8 +30,11 @@ impl Candidate for String {
     }
 }
 
+/// Completion candidate pair
 pub struct Pair {
+    /// Text to display when listing alternatives.
     pub display: String,
+    /// Text to insert in line.
     pub replacement: String,
 }
 
@@ -47,6 +50,7 @@ impl Candidate for Pair {
 
 /// To be called for tab-completion.
 pub trait Completer {
+    /// Specific completion candidate.
     type Candidate: Candidate;
 
     /// Takes the currently edited `line` with the cursor `pos`ition and
@@ -149,14 +153,19 @@ cfg_if::cfg_if! {
     }
 }
 
+/// Kind of quote.
 #[derive(Clone, Copy, Debug, PartialEq)]
 pub enum Quote {
+    /// Double quote: `"`
     Double,
+    /// Single quote: `'`
     Single,
+    /// No quote
     None,
 }
 
 impl FilenameCompleter {
+    /// Constructor
     pub fn new() -> Self {
         Self {
             break_chars: &DEFAULT_BREAK_CHARS,
@@ -164,6 +173,9 @@ impl FilenameCompleter {
         }
     }
 
+    /// Takes the currently edited `line` with the cursor `pos`ition and
+    /// returns the start position and the completion candidates for the
+    /// partial path to be completed.
     pub fn complete_path(&self, line: &str, pos: usize) -> Result<(usize, Vec<Pair>)> {
         let (start, path, esc_char, break_chars, quote) =
             if let Some((idx, quote)) = find_unclosed_quote(&line[..pos]) {
@@ -190,7 +202,9 @@ impl FilenameCompleter {
                 let path = unescape(path, ESCAPE_CHAR);
                 (start, path, ESCAPE_CHAR, &self.break_chars, Quote::None)
             };
-        let matches = filename_complete(&path, esc_char, break_chars, quote)?;
+        let mut matches = filename_complete(&path, esc_char, break_chars, quote)?;
+        #[allow(clippy::unnecessary_sort_by)]
+        matches.sort_by(|a, b| a.display().cmp(b.display()));
         Ok((start, matches))
     }
 }
@@ -285,7 +299,7 @@ fn filename_complete(
     quote: Quote,
 ) -> Result<Vec<Pair>> {
     #[cfg(feature = "with-dirs")]
-    use dirs::home_dir;
+    use dirs_next::home_dir;
     use std::env::current_dir;
 
     let sep = path::MAIN_SEPARATOR;
@@ -332,10 +346,12 @@ fn filename_complete(
 
     // if any of the below IO operations have errors, just ignore them
     if let Ok(read_dir) = dir.read_dir() {
+        let file_name = normalize(file_name);
         for entry in read_dir {
             if let Ok(entry) = entry {
                 if let Some(s) = entry.file_name().to_str() {
-                    if s.starts_with(file_name) {
+                    let ns = normalize(s);
+                    if ns.starts_with(file_name.as_ref()) {
                         if let Ok(metadata) = fs::metadata(entry.path()) {
                             let mut path = String::from(dir_name) + s;
                             if metadata.is_dir() {
@@ -354,6 +370,17 @@ fn filename_complete(
     Ok(entries)
 }
 
+#[cfg(any(windows, target_os = "macos"))]
+fn normalize(s: &str) -> Cow<str> {
+    // case insensitive
+    Cow::Owned(s.to_lowercase())
+}
+
+#[cfg(not(any(windows, target_os = "macos")))]
+fn normalize(s: &str) -> Cow<str> {
+    Cow::Borrowed(s)
+}
+
 /// Given a `line` and a cursor `pos`ition,
 /// try to find backward the start of a word.
 /// Return (0, `line[..pos]`) if no break char has been found.
@@ -393,6 +420,7 @@ pub fn extract_word<'l>(
     }
 }
 
+/// Returns the longest common prefix among all `Candidate::replacement()`s.
 pub fn longest_common_prefix<C: Candidate>(candidates: &[C]) -> Option<&str> {
     if candidates.is_empty() {
         return None;
@@ -540,21 +568,21 @@ mod tests {
 
         let s = "User";
         let c1 = String::from(s);
-        candidates.push(c1.clone());
+        candidates.push(c1);
         {
             let lcp = super::longest_common_prefix(&candidates);
             assert_eq!(Some(s), lcp);
         }
 
         let c2 = String::from("Users");
-        candidates.push(c2.clone());
+        candidates.push(c2);
         {
             let lcp = super::longest_common_prefix(&candidates);
             assert_eq!(Some(s), lcp);
         }
 
         let c3 = String::from("");
-        candidates.push(c3.clone());
+        candidates.push(c3);
         {
             let lcp = super::longest_common_prefix(&candidates);
             assert!(lcp.is_none());
@@ -581,4 +609,10 @@ mod tests {
             super::find_unclosed_quote("\"c:\\users\\All Users\\")
         )
     }
+
+    #[cfg(windows)]
+    #[test]
+    pub fn normalize() {
+        assert_eq!(super::normalize("Windows"), "windows")
+    }
 }
