@@ -89,7 +89,7 @@ impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
                 self.refresh_line()?;
                 continue;
             }
-            if let Ok(Cmd::Replace(_, _)) = rc {
+            if let Ok(Cmd::Replace(..)) = rc {
                 self.changes.borrow_mut().begin();
             }
             return rc;
@@ -142,7 +142,7 @@ impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
     ) -> Result<()> {
         let info = match info {
             Info::NoHint => None,
-            Info::Hint => self.hint.as_ref().map(String::as_str),
+            Info::Hint => self.hint.as_deref(),
             Info::Msg(msg) => msg,
         };
         let highlighter = if self.out.colors_enabled() {
@@ -151,27 +151,9 @@ impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
             None
         };
 
-        // calculate the desired position of the cursor
-        let pos = self.line.pos();
-        let cursor = self.out.calculate_position(&self.line[..pos], prompt_size);
-        // calculate the position of the end of the input line
-        let mut end = if pos == self.line.len() {
-            cursor
-        } else {
-            self.out.calculate_position(&self.line[pos..], cursor)
-        };
-        if let Some(info) = info {
-            end = self.out.calculate_position(&info, end);
-        }
-
-        let new_layout = Layout {
-            prompt_size,
-            default_prompt,
-            cursor,
-            end,
-        };
-        debug_assert!(new_layout.prompt_size <= new_layout.cursor);
-        debug_assert!(new_layout.cursor <= new_layout.end);
+        let new_layout = self
+            .out
+            .compute_layout(prompt_size, default_prompt, &self.line, info);
 
         debug!(target: "rustyline", "old layout: {:?}", self.layout);
         debug!(target: "rustyline", "new layout: {:?}", new_layout);
@@ -225,14 +207,9 @@ impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
             let result = validator.validate(&mut ValidationContext::new(self))?;
             let corrected = self.changes.borrow_mut().end();
             let validated = match result {
-                ValidationResult::Incomplete => {
-                    self.edit_move_end()?;
-                    self.edit_insert('\n', 1)?;
-                    false
-                }
+                ValidationResult::Incomplete => false,
                 ValidationResult::Valid(msg) => {
                     // Accept the line regardless of where the cursor is.
-                    self.edit_move_end()?;
                     if corrected || self.has_hint() || msg.is_some() {
                         // Force a refresh without hints to leave the previous
                         // line as the user typed it after a newline.
@@ -272,12 +249,7 @@ impl<'out, 'prompt, H: Helper> Refresher for State<'out, 'prompt, H> {
         let prompt_size = self.prompt_size;
         self.hint = None;
         self.highlight_char();
-        self.refresh(
-            self.prompt,
-            prompt_size,
-            true,
-            Info::Msg(msg.as_ref().map(String::as_str)),
-        )
+        self.refresh(self.prompt, prompt_size, true, Info::Msg(msg.as_deref()))
     }
 
     fn refresh_prompt_and_line(&mut self, prompt: &str) -> Result<()> {
@@ -464,6 +436,24 @@ impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
         }
     }
 
+    /// Move cursor to the start of the buffer.
+    pub fn edit_move_buffer_start(&mut self) -> Result<()> {
+        if self.line.move_buffer_start() {
+            self.move_cursor()
+        } else {
+            Ok(())
+        }
+    }
+
+    /// Move cursor to the end of the buffer.
+    pub fn edit_move_buffer_end(&mut self) -> Result<()> {
+        if self.line.move_buffer_end() {
+            self.move_cursor()
+        } else {
+            Ok(())
+        }
+    }
+
     pub fn edit_kill(&mut self, mvt: &Movement) -> Result<()> {
         if self.line.kill(mvt) {
             self.refresh_line()
@@ -517,6 +507,26 @@ impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
         }
     }
 
+    /// Moves the cursor to the same column in the line above
+    pub fn edit_move_line_up(&mut self, n: RepeatCount) -> Result<bool> {
+        if self.line.move_to_line_up(n) {
+            self.move_cursor()?;
+            Ok(true)
+        } else {
+            Ok(false)
+        }
+    }
+
+    /// Moves the cursor to the same column in the line above
+    pub fn edit_move_line_down(&mut self, n: RepeatCount) -> Result<bool> {
+        if self.line.move_to_line_down(n) {
+            self.move_cursor()?;
+            Ok(true)
+        } else {
+            Ok(false)
+        }
+    }
+
     pub fn edit_move_to(&mut self, cs: CharSearch, n: RepeatCount) -> Result<()> {
         if self.line.move_to(cs, n) {
             self.move_cursor()
