@@ -14,8 +14,8 @@ use crate::tty::{RawReader, Term, Terminal};
 pub type RepeatCount = usize;
 
 /// Commands
-// #[non_exhaustive]
 #[derive(Debug, Clone, PartialEq)]
+#[non_exhaustive]
 pub enum Cmd {
     /// abort
     Abort, // Miscellaneous Command
@@ -45,7 +45,9 @@ pub enum Cmd {
     HistorySearchBackward,
     /// history-search-forward
     HistorySearchForward,
+    /// Insert text
     Insert(RepeatCount, String),
+    /// Interrupt signal (Ctrl-C)
     Interrupt,
     /// backward-delete-char, backward-kill-line, backward-kill-word
     /// delete-char, kill-line, kill-word, unix-line-discard, unix-word-rubout,
@@ -57,6 +59,7 @@ pub enum Cmd {
     Move(Movement),
     /// next-history
     NextHistory,
+    /// No action
     Noop,
     /// vi-replace
     Overwrite(char),
@@ -72,6 +75,7 @@ pub enum Cmd {
     ReverseSearchHistory,
     /// self-insert
     SelfInsert(RepeatCount, char),
+    /// Suspend signal (Ctrl-Z on unix platform)
     Suspend,
     /// transpose-chars
     TransposeChars,
@@ -79,6 +83,7 @@ pub enum Cmd {
     TransposeWords(RepeatCount),
     /// undo
     Undo(RepeatCount),
+    /// Unsupported / unexpected
     Unknown,
     /// upcase-word
     UpcaseWord,
@@ -88,19 +93,29 @@ pub enum Cmd {
     Yank(RepeatCount, Anchor),
     /// yank-pop
     YankPop,
+    /// moves cursor to the line above or switches to prev history entry if
+    /// the cursor is already on the first line
+    LineUpOrPreviousHistory(RepeatCount),
+    /// moves cursor to the line below or switches to next history entry if
+    /// the cursor is already on the last line
+    LineDownOrNextHistory(RepeatCount),
+    /// accepts the line when cursor is at the end of the text (non including
+    /// trailing whitespace), inserts newline character otherwise
+    AcceptOrInsertLine,
 }
 
 impl Cmd {
+    /// Tells if current command should reset kill ring.
     pub fn should_reset_kill_ring(&self) -> bool {
         #[allow(clippy::match_same_arms)]
         match *self {
             Cmd::Kill(Movement::BackwardChar(_)) | Cmd::Kill(Movement::ForwardChar(_)) => true,
             Cmd::ClearScreen
             | Cmd::Kill(_)
-            | Cmd::Replace(_, _)
+            | Cmd::Replace(..)
             | Cmd::Noop
             | Cmd::Suspend
-            | Cmd::Yank(_, _)
+            | Cmd::Yank(..)
             | Cmd::YankPop => false,
             _ => true,
         }
@@ -108,15 +123,15 @@ impl Cmd {
 
     fn is_repeatable_change(&self) -> bool {
         match *self {
-            Cmd::Insert(_, _)
+            Cmd::Insert(..)
             | Cmd::Kill(_)
-            | Cmd::ReplaceChar(_, _)
-            | Cmd::Replace(_, _)
-            | Cmd::SelfInsert(_, _)
+            | Cmd::ReplaceChar(..)
+            | Cmd::Replace(..)
+            | Cmd::SelfInsert(..)
             | Cmd::ViYankTo(_)
-            | Cmd::Yank(_, _) => true,
-            Cmd::TransposeChars // TODO Validate
-            | _ => false,
+            | Cmd::Yank(..) => true,
+            // Cmd::TransposeChars | TODO Validate
+            _ => false,
         }
     }
 
@@ -188,26 +203,33 @@ pub enum Word {
 /// Where to move with respect to word boundary
 #[derive(Debug, Clone, PartialEq, Copy)]
 pub enum At {
+    /// Start of word.
     Start,
+    /// Before end of word.
     BeforeEnd,
+    /// After end of word.
     AfterEnd,
 }
 
 /// Where to paste (relative to cursor position)
 #[derive(Debug, Clone, PartialEq, Copy)]
 pub enum Anchor {
+    /// After cursor
     After,
+    /// Before cursor
     Before,
 }
 
 /// Vi character search
 #[derive(Debug, Clone, PartialEq, Copy)]
 pub enum CharSearch {
+    /// Forward search
     Forward(char),
-    // until
+    /// Forward search until
     ForwardBefore(char),
+    /// Backward search
     Backward(char),
-    // until
+    /// Backward search until
     BackwardAfter(char),
 }
 
@@ -224,8 +246,10 @@ impl CharSearch {
 
 /// Where to move
 #[derive(Debug, Clone, PartialEq)]
+#[non_exhaustive]
 pub enum Movement {
-    WholeLine, // not really a movement
+    /// Whole current line (not really a movement but a range)
+    WholeLine,
     /// beginning-of-line
     BeginningOfLine,
     /// end-of-line
@@ -242,6 +266,16 @@ pub enum Movement {
     BackwardChar(RepeatCount),
     /// forward-char
     ForwardChar(RepeatCount),
+    /// move to the same column on the previous line
+    LineUp(RepeatCount),
+    /// move to the same column on the next line
+    LineDown(RepeatCount),
+    /// Whole user input (not really a movement but a range)
+    WholeBuffer,
+    /// beginning-of-buffer
+    BeginningOfBuffer,
+    /// end-of-buffer
+    EndOfBuffer,
 }
 
 impl Movement {
@@ -263,6 +297,11 @@ impl Movement {
             }
             Movement::BackwardChar(previous) => Movement::BackwardChar(repeat_count(previous, new)),
             Movement::ForwardChar(previous) => Movement::ForwardChar(repeat_count(previous, new)),
+            Movement::LineUp(previous) => Movement::LineUp(repeat_count(previous, new)),
+            Movement::LineDown(previous) => Movement::LineDown(repeat_count(previous, new)),
+            Movement::WholeBuffer => Movement::WholeBuffer,
+            Movement::BeginningOfBuffer => Movement::BeginningOfBuffer,
+            Movement::EndOfBuffer => Movement::EndOfBuffer,
         }
     }
 }
@@ -342,9 +381,18 @@ impl InputState {
         single_esc_abort: bool,
     ) -> Result<Cmd> {
         match self.mode {
-            EditMode::Emacs => self.emacs(rdr, wrt, single_esc_abort),
-            EditMode::Vi if self.input_mode != InputMode::Command => self.vi_insert(rdr, wrt),
-            EditMode::Vi => self.vi_command(rdr, wrt),
+            EditMode::Emacs => {
+                let key = rdr.next_key(single_esc_abort)?;
+                self.emacs(rdr, wrt, key)
+            }
+            EditMode::Vi if self.input_mode != InputMode::Command => {
+                let key = rdr.next_key(false)?;
+                self.vi_insert(rdr, wrt, key)
+            }
+            EditMode::Vi => {
+                let key = rdr.next_key(false)?;
+                self.vi_command(rdr, wrt, key)
+            }
         }
     }
 
@@ -393,9 +441,8 @@ impl InputState {
         &mut self,
         rdr: &mut R,
         wrt: &mut dyn Refresher,
-        single_esc_abort: bool,
+        mut key: KeyPress,
     ) -> Result<Cmd> {
-        let mut key = rdr.next_key(single_esc_abort)?;
         if let KeyPress::Meta(digit @ '-') = key {
             key = self.emacs_digit_argument(rdr, wrt, digit)?;
         } else if let KeyPress::Meta(digit @ '0'..='9') = key {
@@ -540,8 +587,12 @@ impl InputState {
         }
     }
 
-    fn vi_command<R: RawReader>(&mut self, rdr: &mut R, wrt: &mut dyn Refresher) -> Result<Cmd> {
-        let mut key = rdr.next_key(false)?;
+    fn vi_command<R: RawReader>(
+        &mut self,
+        rdr: &mut R,
+        wrt: &mut dyn Refresher,
+        mut key: KeyPress,
+    ) -> Result<Cmd> {
         if let KeyPress::Char(digit @ '1'..='9') = key {
             key = self.vi_arg_digit(rdr, wrt, digit)?;
         }
@@ -563,16 +614,17 @@ impl InputState {
             }
         }
         let cmd = match key {
-            KeyPress::Char('$') |
-            KeyPress::End => Cmd::Move(Movement::EndOfLine),
-            KeyPress::Char('.') => { // vi-redo (repeat last command)
+            KeyPress::Char('$') | KeyPress::End => Cmd::Move(Movement::EndOfLine),
+            KeyPress::Char('.') => {
+                // vi-redo (repeat last command)
                 if no_num_args {
                     self.last_cmd.redo(None, wrt)
                 } else {
                     self.last_cmd.redo(Some(n), wrt)
                 }
-            },
-            // TODO KeyPress::Char('%') => Cmd::???, Move to the corresponding opening/closing bracket
+            }
+            // TODO KeyPress::Char('%') => Cmd::???, Move to the corresponding opening/closing
+            // bracket
             KeyPress::Char('0') => Cmd::Move(Movement::BeginningOfLine),
             KeyPress::Char('^') => Cmd::Move(Movement::ViFirstPrint),
             KeyPress::Char('a') => {
@@ -600,14 +652,11 @@ impl InputState {
                 self.input_mode = InputMode::Insert;
                 Cmd::Replace(Movement::EndOfLine, None)
             }
-            KeyPress::Char('d') => {
-                match self.vi_cmd_motion(rdr, wrt, key, n)? {
-                    Some(mvt) => Cmd::Kill(mvt),
-                    None => Cmd::Unknown,
-                }
-            }
-            KeyPress::Char('D') |
-            KeyPress::Ctrl('K') => Cmd::Kill(Movement::EndOfLine),
+            KeyPress::Char('d') => match self.vi_cmd_motion(rdr, wrt, key, n)? {
+                Some(mvt) => Cmd::Kill(mvt),
+                None => Cmd::Unknown,
+            },
+            KeyPress::Char('D') | KeyPress::Ctrl('K') => Cmd::Kill(Movement::EndOfLine),
             KeyPress::Char('e') => Cmd::Move(Movement::ForwardWord(n, At::BeforeEnd, Word::Vi)),
             KeyPress::Char('E') => Cmd::Move(Movement::ForwardWord(n, At::BeforeEnd, Word::Big)),
             KeyPress::Char('i') => {
@@ -630,18 +679,14 @@ impl InputState {
                     None => Cmd::Unknown,
                 }
             }
-            KeyPress::Char(';') => {
-                match self.last_char_search {
-                    Some(cs) => Cmd::Move(Movement::ViCharSearch(n, cs)),
-                    None => Cmd::Noop,
-                }
-            }
-            KeyPress::Char(',') => {
-                match self.last_char_search {
-                    Some(ref cs) => Cmd::Move(Movement::ViCharSearch(n, cs.opposite())),
-                    None => Cmd::Noop,
-                }
-            }
+            KeyPress::Char(';') => match self.last_char_search {
+                Some(cs) => Cmd::Move(Movement::ViCharSearch(n, cs)),
+                None => Cmd::Noop,
+            },
+            KeyPress::Char(',') => match self.last_char_search {
+                Some(ref cs) => Cmd::Move(Movement::ViCharSearch(n, cs.opposite())),
+                None => Cmd::Noop,
+            },
             // TODO KeyPress::Char('G') => Cmd::???, Move to the history line n
             KeyPress::Char('p') => Cmd::Yank(n, Anchor::After), // vi-put
             KeyPress::Char('P') => Cmd::Yank(n, Anchor::Before), // vi-put
@@ -671,29 +716,27 @@ impl InputState {
             }
             KeyPress::Char('u') => Cmd::Undo(n),
             // KeyPress::Char('U') => Cmd::???, // revert-line
-            KeyPress::Char('w') => Cmd::Move(Movement::ForwardWord(n, At::Start, Word::Vi)), // vi-next-word
-            KeyPress::Char('W') => Cmd::Move(Movement::ForwardWord(n, At::Start, Word::Big)), // vi-next-word
-            KeyPress::Char('x') => Cmd::Kill(Movement::ForwardChar(n)), // vi-delete: TODO move backward if eol
+            KeyPress::Char('w') => Cmd::Move(Movement::ForwardWord(n, At::Start, Word::Vi)), /* vi-next-word */
+            KeyPress::Char('W') => Cmd::Move(Movement::ForwardWord(n, At::Start, Word::Big)), /* vi-next-word */
+            // TODO move backward if eol
+            KeyPress::Char('x') => Cmd::Kill(Movement::ForwardChar(n)), // vi-delete
             KeyPress::Char('X') => Cmd::Kill(Movement::BackwardChar(n)), // vi-rubout
-            KeyPress::Char('y') => {
-                match self.vi_cmd_motion(rdr, wrt, key, n)? {
-                    Some(mvt) => Cmd::ViYankTo(mvt),
-                    None => Cmd::Unknown,
-                }
-            }
+            KeyPress::Char('y') => match self.vi_cmd_motion(rdr, wrt, key, n)? {
+                Some(mvt) => Cmd::ViYankTo(mvt),
+                None => Cmd::Unknown,
+            },
             // KeyPress::Char('Y') => Cmd::???, // vi-yank-to
-            KeyPress::Char('h') |
-            KeyPress::Ctrl('H') |
-            KeyPress::Backspace => Cmd::Move(Movement::BackwardChar(n)),
+            KeyPress::Char('h') | KeyPress::Ctrl('H') | KeyPress::Backspace => {
+                Cmd::Move(Movement::BackwardChar(n))
+            }
             KeyPress::Ctrl('G') => Cmd::Abort,
-            KeyPress::Char('l') |
-            KeyPress::Char(' ') => Cmd::Move(Movement::ForwardChar(n)),
+            KeyPress::Char('l') | KeyPress::Char(' ') => Cmd::Move(Movement::ForwardChar(n)),
             KeyPress::Ctrl('L') => Cmd::ClearScreen,
-            KeyPress::Char('+') |
-            KeyPress::Char('j') | // TODO: move to the start of the line.
+            KeyPress::Char('+') | KeyPress::Char('j') => Cmd::LineDownOrNextHistory(n),
+            // TODO: move to the start of the line.
             KeyPress::Ctrl('N') => Cmd::NextHistory,
-            KeyPress::Char('-') |
-            KeyPress::Char('k') | // TODO: move to the start of the line.
+            KeyPress::Char('-') | KeyPress::Char('k') => Cmd::LineUpOrPreviousHistory(n),
+            // TODO: move to the start of the line.
             KeyPress::Ctrl('P') => Cmd::PreviousHistory,
             KeyPress::Ctrl('R') => {
                 self.input_mode = InputMode::Insert; // TODO Validate
@@ -713,8 +756,12 @@ impl InputState {
         Ok(cmd)
     }
 
-    fn vi_insert<R: RawReader>(&mut self, rdr: &mut R, wrt: &mut dyn Refresher) -> Result<Cmd> {
-        let key = rdr.next_key(false)?;
+    fn vi_insert<R: RawReader>(
+        &mut self,
+        rdr: &mut R,
+        wrt: &mut dyn Refresher,
+        key: KeyPress,
+    ) -> Result<Cmd> {
         {
             let bindings = self.custom_bindings.read().unwrap();
             if let Some(cmd) = bindings.get(&key) {
@@ -739,6 +786,13 @@ impl InputState {
             KeyPress::Tab => Cmd::Complete,
             // Don't complete hints when the cursor is not at the end of a line
             KeyPress::Right if wrt.has_hint() && wrt.is_cursor_at_end() => Cmd::CompleteHint,
+            KeyPress::Meta(k) => {
+                debug!(target: "rustyline", "Vi fast command mode: {}", k);
+                self.input_mode = InputMode::Command;
+                wrt.done_inserting();
+
+                self.vi_command(rdr, wrt, KeyPress::Char(k))?
+            }
             KeyPress::Esc => {
                 // vi-movement-mode/vi-command-mode
                 self.input_mode = InputMode::Command;
@@ -749,9 +803,9 @@ impl InputState {
         };
         debug!(target: "rustyline", "Vi insert: {:?}", cmd);
         if cmd.is_repeatable_change() {
-            if let (Cmd::Replace(_, _), Cmd::SelfInsert(_, _)) = (&self.last_cmd, &cmd) {
+            if let (Cmd::Replace(..), Cmd::SelfInsert(..)) = (&self.last_cmd, &cmd) {
                 // replacing...
-            } else if let (Cmd::SelfInsert(_, _), Cmd::SelfInsert(_, _)) = (&self.last_cmd, &cmd) {
+            } else if let (Cmd::SelfInsert(..), Cmd::SelfInsert(..)) = (&self.last_cmd, &cmd) {
                 // inserting...
             } else {
                 self.last_cmd = cmd.clone();
@@ -804,6 +858,8 @@ impl InputState {
                 Some(Movement::BackwardChar(n))
             }
             KeyPress::Char('l') | KeyPress::Char(' ') => Some(Movement::ForwardChar(n)),
+            KeyPress::Char('j') | KeyPress::Char('+') => Some(Movement::LineDown(n)),
+            KeyPress::Char('k') | KeyPress::Char('-') => Some(Movement::LineUp(n)),
             KeyPress::Char('w') => {
                 // 'cw' is 'ce'
                 if key == KeyPress::Char('c') {
@@ -881,8 +937,8 @@ impl InputState {
             }
             KeyPress::Ctrl('J') |
             KeyPress::Enter => Cmd::AcceptLine,
-            KeyPress::Down => Cmd::NextHistory,
-            KeyPress::Up => Cmd::PreviousHistory,
+            KeyPress::Down => Cmd::LineDownOrNextHistory(1),
+            KeyPress::Up => Cmd::LineUpOrPreviousHistory(1),
             KeyPress::Ctrl('R') => Cmd::ReverseSearchHistory,
             KeyPress::Ctrl('S') => Cmd::ForwardSearchHistory, // most terminals override Ctrl+S to suspend execution
             KeyPress::Ctrl('T') => Cmd::TransposeChars,
