@@ -15,7 +15,7 @@
 //!     Err(_) => println!("No input"),
 //! }
 //! ```
-// #![feature(non_exhaustive)]
+#![warn(missing_docs)]
 
 pub mod completion;
 pub mod config;
@@ -70,6 +70,9 @@ fn complete_line<H: Helper>(
     input_state: &mut InputState,
     config: &Config,
 ) -> Result<Option<Cmd>> {
+    #[cfg(all(unix, feature = "with-fuzzy"))]
+    use skim::{Skim, SkimOptionsBuilder};
+
     let completer = s.helper.unwrap();
     // get a list of completions
     let (start, candidates) = completer.complete(&s.line, s.line.pos(), &s.ctx)?;
@@ -185,6 +188,44 @@ fn complete_line<H: Helper>(
             Ok(None)
         }
     } else {
+        // if fuzzy feature is enabled and on unix based systems check for the
+        // corresponding completion_type
+        #[cfg(all(unix, feature = "with-fuzzy"))]
+        {
+            if CompletionType::Fuzzy == config.completion_type() {
+                // skim takes input of candidates separated by new line
+                let input = candidates
+                    .iter()
+                    .map(|c| c.display())
+                    .collect::<Vec<_>>()
+                    .join("\n");
+
+                // setup skim and run with input options
+                // will display UI for fuzzy search and return selected results
+                // by default skim multi select is off so only expect one selection
+
+                let options = SkimOptionsBuilder::default()
+                    .height(Some("20%"))
+                    .prompt(Some("? "))
+                    .reverse(true)
+                    .build()
+                    .unwrap();
+
+                let selected_items =
+                    Skim::run_with(&options, Some(Box::new(std::io::Cursor::new(input))))
+                        .map(|out| out.selected_items)
+                        .unwrap_or_else(Vec::new);
+
+                // match the first (and only) returned option with the candidate and update the
+                // line otherwise only refresh line to clear the skim UI changes
+                if let Some(item) = selected_items.first() {
+                    if let Some(candidate) = candidates.get(item.get_index()) {
+                        completer.update(&mut s.line, start, candidate.replacement());
+                    }
+                }
+                s.refresh_line()?;
+            }
+        };
         Ok(None)
     }
 }
@@ -240,6 +281,7 @@ fn page_completions<C: Candidate, H: Helper>(
                 && cmd != Cmd::SelfInsert(1, ' ')
                 && cmd != Cmd::Kill(Movement::BackwardChar(1))
                 && cmd != Cmd::AcceptLine
+                && cmd != Cmd::AcceptOrInsertLine
             {
                 cmd = s.next_cmd(input_state, rdr, false)?;
             }
@@ -247,7 +289,7 @@ fn page_completions<C: Candidate, H: Helper>(
                 Cmd::SelfInsert(1, 'y') | Cmd::SelfInsert(1, 'Y') | Cmd::SelfInsert(1, ' ') => {
                     pause_row += s.out.get_rows() - 1;
                 }
-                Cmd::AcceptLine => {
+                Cmd::AcceptLine | Cmd::AcceptOrInsertLine => {
                     pause_row += 1;
                 }
                 _ => break,
@@ -277,6 +319,8 @@ fn page_completions<C: Candidate, H: Helper>(
         s.out.write_and_flush(ab.as_bytes())?;
     }
     s.out.write_and_flush(b"\n")?;
+    s.layout.end.row = 0; // dirty way to make clear_old_rows do nothing
+    s.layout.cursor.row = 0;
     s.refresh_line()?;
     Ok(None)
 }
@@ -395,7 +439,13 @@ fn readline_edit<H: Helper>(
 
     let mut rdr = editor.term.create_reader(&editor.config)?;
     if editor.term.is_output_tty() {
-        s.move_cursor_at_leftmost(&mut rdr)?;
+        if let Err(e) = s.move_cursor_at_leftmost(&mut rdr) {
+            if s.out.sigwinch() {
+                s.out.update_size();
+            } else {
+                return Err(e);
+            }
+        }
     }
     s.refresh_line()?;
 
@@ -417,7 +467,7 @@ fn readline_edit<H: Helper>(
             }
         }
 
-        if let Cmd::CompleteHint = cmd {
+        if Cmd::CompleteHint == cmd {
             complete_hint_line(&mut s)?;
             continue;
         }
@@ -495,6 +545,16 @@ fn readline_edit<H: Helper>(
                 // Fetch the previous command from the history list.
                 s.edit_history_next(true)?
             }
+            Cmd::LineUpOrPreviousHistory(n) => {
+                if !s.edit_move_line_up(n)? {
+                    s.edit_history_next(true)?
+                }
+            }
+            Cmd::LineDownOrNextHistory(n) => {
+                if !s.edit_move_line_down(n)? {
+                    s.edit_history_next(false)?
+                }
+            }
             Cmd::HistorySearchBackward => s.edit_history_search(Direction::Reverse)?,
             Cmd::HistorySearchForward => s.edit_history_search(Direction::Forward)?,
             Cmd::TransposeChars => {
@@ -521,20 +581,21 @@ fn readline_edit<H: Helper>(
                     kill_ring.kill(&text, Mode::Append)
                 }
             }
-            Cmd::AcceptLine => {
+            Cmd::AcceptLine | Cmd::AcceptOrInsertLine => {
                 #[cfg(test)]
                 {
                     editor.term.cursor = s.layout.cursor.col;
                 }
-                // Accept the line regardless of where the cursor is.
                 if s.has_hint() || !s.is_default_prompt() {
                     // Force a refresh without hints to leave the previous
                     // line as the user typed it after a newline.
                     s.refresh_line_with_msg(None)?;
                 }
-                s.edit_move_end()?;
-                if s.validate()? {
+                // Only accept value if cursor is at the end of the buffer
+                if s.validate()? && (cmd == Cmd::AcceptLine || s.line.is_end_of_input()) {
                     break;
+                } else {
+                    s.edit_insert('\n', 1)?;
                 }
                 continue;
             }
@@ -561,6 +622,20 @@ fn readline_edit<H: Helper>(
                 // move forwards one word
                 s.edit_move_to_next_word(at, word_def, n)?
             }
+            Cmd::Move(Movement::LineUp(n)) => {
+                s.edit_move_line_up(n)?;
+            }
+            Cmd::Move(Movement::LineDown(n)) => {
+                s.edit_move_line_down(n)?;
+            }
+            Cmd::Move(Movement::BeginningOfBuffer) => {
+                // Move to the start of the buffer.
+                s.edit_move_buffer_start()?
+            }
+            Cmd::Move(Movement::EndOfBuffer) => {
+                // Move to the end of the buffer.
+                s.edit_move_buffer_end()?
+            }
             Cmd::DowncaseWord => {
                 // lowercase word after point
                 s.edit_word(WordAction::LOWERCASE)?
@@ -597,7 +672,7 @@ fn readline_edit<H: Helper>(
                 s.refresh_line()?;
                 continue;
             }
-            Cmd::Noop | _ => {
+            _ => {
                 // Ignore the character typed.
             }
         }
@@ -828,6 +903,7 @@ impl<H: Helper> Editor<H> {
         }
     }
 
+    /// Returns an iterator over edited lines
     /// ```
     /// let mut rl = rustyline::Editor::<()>::new();
     /// for readline in rl.iter("> ") {
@@ -842,7 +918,7 @@ impl<H: Helper> Editor<H> {
     ///     }
     /// }
     /// ```
-    pub fn iter<'a>(&'a mut self, prompt: &'a str) -> Iter<'_, H> {
+    pub fn iter<'a>(&'a mut self, prompt: &'a str) -> impl Iterator<Item = Result<String>> + 'a {
         Iter {
             editor: self,
             prompt,
@@ -901,8 +977,7 @@ impl<H: Helper> fmt::Debug for Editor<H> {
     }
 }
 
-/// Edited lines iterator
-pub struct Iter<'a, H: Helper> {
+struct Iter<'a, H: Helper> {
     editor: &'a mut Editor<H>,
     prompt: &'a str,
 }
@@ -925,3 +1000,6 @@ impl<'a, H: Helper> Iterator for Iter<'a, H> {
 extern crate assert_matches;
 #[cfg(test)]
 mod test;
+
+#[cfg(doctest)]
+doc_comment::doctest!("../README.md");
