@@ -1,4 +1,7 @@
 //! This module implements and describes common TTY methods & traits
+
+use unicode_width::UnicodeWidthStr;
+
 use crate::config::{BellStyle, ColorMode, Config, OutputStreamType};
 use crate::highlight::Highlighter;
 use crate::keys::KeyPress;
@@ -41,6 +44,40 @@ pub trait Renderer {
         highlighter: Option<&dyn Highlighter>,
     ) -> Result<()>;
 
+    /// Compute layout for rendering prompt + line + some info (either hint,
+    /// validation msg, ...). on the screen. Depending on screen width, line
+    /// wrapping may be applied.
+    fn compute_layout(
+        &self,
+        prompt_size: Position,
+        default_prompt: bool,
+        line: &LineBuffer,
+        info: Option<&str>,
+    ) -> Layout {
+        // calculate the desired position of the cursor
+        let pos = line.pos();
+        let cursor = self.calculate_position(&line[..pos], prompt_size);
+        // calculate the position of the end of the input line
+        let mut end = if pos == line.len() {
+            cursor
+        } else {
+            self.calculate_position(&line[pos..], cursor)
+        };
+        if let Some(info) = info {
+            end = self.calculate_position(&info, end);
+        }
+
+        let new_layout = Layout {
+            prompt_size,
+            default_prompt,
+            cursor,
+            end,
+        };
+        debug_assert!(new_layout.prompt_size <= new_layout.cursor);
+        debug_assert!(new_layout.cursor <= new_layout.end);
+        new_layout
+    }
+
     /// Calculate the number of columns and rows used to display `s` on a
     /// `cols` width terminal starting at `orig`.
     fn calculate_position(&self, s: &str, orig: Position) -> Position;
@@ -129,6 +166,37 @@ impl<'a, R: Renderer + ?Sized> Renderer for &'a mut R {
     }
 }
 
+// ignore ANSI escape sequence
+fn width(s: &str, esc_seq: &mut u8) -> usize {
+    if *esc_seq == 1 {
+        if s == "[" {
+            // CSI
+            *esc_seq = 2;
+        } else {
+            // two-character sequence
+            *esc_seq = 0;
+        }
+        0
+    } else if *esc_seq == 2 {
+        if s == ";" || (s.as_bytes()[0] >= b'0' && s.as_bytes()[0] <= b'9') {
+            /*} else if s == "m" {
+            // last
+             *esc_seq = 0;*/
+        } else {
+            // not supported
+            *esc_seq = 0;
+        }
+        0
+    } else if s == "\x1b" {
+        *esc_seq = 1;
+        0
+    } else if s == "\n" {
+        0
+    } else {
+        s.width()
+    }
+}
+
 /// Terminal contract
 pub trait Term {
     type Reader: RawReader; // rl_instream
@@ -156,19 +224,21 @@ pub trait Term {
     fn create_writer(&self) -> Self::Writer;
 }
 
-cfg_if::cfg_if! {
-    if #[cfg(any(test, target_arch = "wasm32"))] {
-        mod test;
-        pub use self::test::*;
-    } else if #[cfg(windows)] {
-        // If on Windows platform import Windows TTY module
-        // and re-export into mod.rs scope
-        mod windows;
-        pub use self::windows::*;
-    } else if #[cfg(unix)] {
-        // If on Unix platform import Unix TTY module
-        // and re-export into mod.rs scope
-        mod unix;
-        pub use self::unix::*;
-    }
-}
+// If on Windows platform import Windows TTY module
+// and re-export into mod.rs scope
+#[cfg(all(windows, not(target_arch = "wasm32")))]
+mod windows;
+#[cfg(all(windows, not(target_arch = "wasm32")))]
+pub use self::windows::*;
+
+// If on Unix platform import Unix TTY module
+// and re-export into mod.rs scope
+#[cfg(all(unix, not(target_arch = "wasm32")))]
+mod unix;
+#[cfg(all(unix, not(target_arch = "wasm32")))]
+pub use self::unix::*;
+
+#[cfg(any(test, target_arch = "wasm32"))]
+mod test;
+#[cfg(any(test, target_arch = "wasm32"))]
+pub use self::test::*;
