@@ -1,23 +1,19 @@
 //! Unix specific definitions
-use std;
 use std::cmp::Ordering;
 use std::io::{self, Read, Write};
 use std::os::unix::io::{AsRawFd, RawFd};
 use std::sync;
 use std::sync::atomic;
 
-use libc;
 use log::{debug, warn};
-use nix;
 use nix::poll::{self, PollFlags};
 use nix::sys::signal;
 use nix::sys::termios;
 use nix::sys::termios::SetArg;
 use unicode_segmentation::UnicodeSegmentation;
-use unicode_width::UnicodeWidthStr;
 use utf8parse::{Parser, Receiver};
 
-use super::{RawMode, RawReader, Renderer, Term};
+use super::{width, RawMode, RawReader, Renderer, Term};
 use crate::config::{BellStyle, ColorMode, Config, OutputStreamType};
 use crate::error;
 use crate::highlight::Highlighter;
@@ -45,14 +41,34 @@ impl AsRawFd for OutputStreamType {
 
 nix::ioctl_read_bad!(win_size, libc::TIOCGWINSZ, libc::winsize);
 
-#[allow(clippy::identity_conversion)]
+#[allow(clippy::useless_conversion)]
 fn get_win_size<T: AsRawFd + ?Sized>(fileno: &T) -> (usize, usize) {
     use std::mem::zeroed;
 
+    if cfg!(test) {
+        return (80, 24);
+    }
+
     unsafe {
         let mut size: libc::winsize = zeroed();
         match win_size(fileno.as_raw_fd(), &mut size) {
-            Ok(0) => (size.ws_col as usize, size.ws_row as usize), // TODO getCursorPosition
+            Ok(0) => {
+                // In linux pseudo-terminals are created with dimensions of
+                // zero. If host application didn't initialize the correct
+                // size before start we treat zero size as 80 columns and
+                // inifinite rows
+                let cols = if size.ws_col == 0 {
+                    80
+                } else {
+                    size.ws_col as usize
+                };
+                let rows = if size.ws_row == 0 {
+                    usize::max_value()
+                } else {
+                    size.ws_row as usize
+                };
+                (cols, rows)
+            }
             _ => (80, 24),
         }
     }
@@ -484,6 +500,25 @@ impl PosixRenderer {
             bell_style,
         }
     }
+
+    fn clear_old_rows(&mut self, layout: &Layout) {
+        use std::fmt::Write;
+        let current_row = layout.cursor.row;
+        let old_rows = layout.end.row;
+        // old_rows < cursor_row if the prompt spans multiple lines and if
+        // this is the default State.
+        let cursor_row_movement = old_rows.saturating_sub(current_row);
+        // move the cursor down as required
+        if cursor_row_movement > 0 {
+            write!(self.buffer, "\x1b[{}B", cursor_row_movement).unwrap();
+        }
+        // clear old rows
+        for _ in 0..old_rows {
+            self.buffer.push_str("\r\x1b[0K\x1b[A");
+        }
+        // clear the line
+        self.buffer.push_str("\r\x1b[0K");
+    }
 }
 
 impl Renderer for PosixRenderer {
@@ -546,22 +581,8 @@ impl Renderer for PosixRenderer {
         let default_prompt = new_layout.default_prompt;
         let cursor = new_layout.cursor;
         let end_pos = new_layout.end;
-        let current_row = old_layout.cursor.row;
-        let old_rows = old_layout.end.row;
 
-        // old_rows < cursor.row if the prompt spans multiple lines and if
-        // this is the default State.
-        let cursor_row_movement = old_rows.saturating_sub(current_row);
-        // move the cursor down as required
-        if cursor_row_movement > 0 {
-            write!(self.buffer, "\x1b[{}B", cursor_row_movement).unwrap();
-        }
-        // clear old rows
-        for _ in 0..old_rows {
-            self.buffer.push_str("\r\x1b[0K\x1b[A");
-        }
-        // clear the line
-        self.buffer.push_str("\r\x1b[0K");
+        self.clear_old_rows(old_layout);
 
         if let Some(highlighter) = highlighter {
             // display the prompt
@@ -706,36 +727,6 @@ impl Renderer for PosixRenderer {
     }
 }
 
-fn width(s: &str, esc_seq: &mut u8) -> usize {
-    if *esc_seq == 1 {
-        if s == "[" {
-            // CSI
-            *esc_seq = 2;
-        } else {
-            // two-character sequence
-            *esc_seq = 0;
-        }
-        0
-    } else if *esc_seq == 2 {
-        if s == ";" || (s.as_bytes()[0] >= b'0' && s.as_bytes()[0] <= b'9') {
-            /*} else if s == "m" {
-            // last
-             *esc_seq = 0;*/
-        } else {
-            // not supported
-            *esc_seq = 0;
-        }
-        0
-    } else if s == "\x1b" {
-        *esc_seq = 1;
-        0
-    } else if s == "\n" {
-        0
-    } else {
-        s.width()
-    }
-}
-
 fn read_digits_until(rdr: &mut PosixRawReader, sep: char) -> Result<Option<u32>> {
     let mut num: u32 = 0;
     loop {
@@ -923,6 +914,7 @@ fn write_and_flush(out: OutputStreamType, buf: &[u8]) -> Result<()> {
 mod test {
     use super::{Position, PosixRenderer, PosixTerminal, Renderer};
     use crate::config::{BellStyle, OutputStreamType};
+    use crate::line_buffer::LineBuffer;
 
     #[test]
     #[ignore]
@@ -953,4 +945,29 @@ mod test {
         fn assert_sync<T: Sync>() {}
         assert_sync::<PosixTerminal>();
     }
+
+    #[test]
+    fn test_line_wrap() {
+        let mut out = PosixRenderer::new(OutputStreamType::Stdout, 4, true, BellStyle::default());
+        let prompt = "> ";
+        let default_prompt = true;
+        let prompt_size = out.calculate_position(prompt, Position::default());
+
+        let mut line = LineBuffer::init("", 0, None);
+        let old_layout = out.compute_layout(prompt_size, default_prompt, &line, None);
+        assert_eq!(Position { col: 2, row: 0 }, old_layout.cursor);
+        assert_eq!(old_layout.cursor, old_layout.end);
+
+        assert_eq!(Some(true), line.insert('a', out.cols - prompt_size.col + 1));
+        let new_layout = out.compute_layout(prompt_size, default_prompt, &line, None);
+        assert_eq!(Position { col: 1, row: 1 }, new_layout.cursor);
+        assert_eq!(new_layout.cursor, new_layout.end);
+        out.refresh_line(prompt, &line, None, &old_layout, &new_layout, None)
+            .unwrap();
+        #[rustfmt::skip]
+        assert_eq!(
+            "\r\u{1b}[0K> aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\r\u{1b}[1C",
+            out.buffer
+        );
+    }
 }
